---
title: llarp/rpc/rpc_request.hpp

---

# llarp/rpc/rpc_request.hpp



## Namespaces

| Name           |
| -------------- |
| **[llarp](Namespaces/namespacellarp.md)** <br>[crypto.hpp] |
| **[llarp::rpc](Namespaces/namespacellarp_1_1rpc.md)**  |

## Classes

|                | Name           |
| -------------- | -------------- |
| struct | **[llarp::rpc::rpc_error](Classes/structllarp_1_1rpc_1_1rpc__error.md)**  |




## Source code

```cpp
#pragma once

#include "json_bt.hpp"
#include <string_view>
#include <llarp/config/config.hpp>
#include <oxenmq/oxenmq.h>
#include <oxenmq/address.h>
#include <oxen/log/omq_logger.hpp>


namespace llarp::rpc {

  using nlohmann::json;

  using rpc_input = std::variant<nlohmann::json, oxenc::bt_dict_consumer>;

  //  Exception used to signal various types of errors with a request back to the caller.  This
  //  exception indicates that the caller did something wrong: bad data, invalid value, etc., but
  //  don't indicate a local problem (and so we'll log them only at debug).  For more serious,
  //  internal errors a command should throw some other stl error (e.g. std::runtime_error or
  //  perhaps std::logic_error), which will result in a local daemon warning (and a generic internal
  //  error response to the user).
  // 
  //  For JSON RPC these become an error response with the code as the error.code value and the
  //  string as the error.message.
  //  For HTTP JSON these become a 500 Internal Server Error response with the message as the body.
  //  For OxenMQ the code becomes the first part of the response and the message becomes the
  //  second part of the response.
  struct rpc_error : std::runtime_error {
    rpc_error(int16_t code, std::string message)
      : std::runtime_error{"RPC error " + std::to_string(code) + ": " + message},
        code{code}, message{std::move(message)} {}

    int16_t code;
    std::string message;
  };

  template <typename RPC, typename RPCServer, typename RPCCallback>
  void make_invoke() {
    return [](oxenmq::Message& m, RPCServer& server) -> typename RPCCallback::result_type {
      RPC rpc{};    // endpoint struct

      if (m.data.size() > 1)
        m.send_reply(CreateJSONError("Bad Request: RPC requests must have at most one data part "
          "(received " + std::to_string(m.data.size()) + ")"));

      std::string_view request;
      if (!m.data.empty())
        request = m.data[0];

      // parsing input as bt or json
      //    hand off to parse_request (overloaded versions)
      try {
          if (request.front() == 'd') { // Looks like a bt dict
            rpc.set_bt();
            parse_request(rpc, oxenc::bt_dict_consumer{*request});
          }
          else
            parse_request(rpc, m.data.empty() ? json::object() : json::parse(m.data));
      } catch (const std::exception& e) {
        m.send_reply(CreateJSONError("Failed to parse request parameters: "s + e.what()));
      }

      try {
        server.invoke(rpc, std::move(request));
      } catch (const rpc_error& e) {
        log::info(logcat, "RPC request 'rpc.{}' failed with: {}", name, e.what());
        rpc.response = CreateJSONError(fmt::format("RPC request 'rpc.{}' failed with: {}", name, e.what()));
      } catch (const std::exception& e) {
        log::info(logcat, "RPC request 'rpc.{}' raised an exception: {}", name, e.what());
        rpc.response = CreateJSONError(fmt::format("RPC request 'rpc.{}' raised an exception: {}", name, e.what()));
      };

      if (rpc.response.is_null())
        rpc.response = json::object();

      auto _response = (rpc.is_bt()) ? oxenc::bt_serialize(json_to_bt(rpc.response)) : rpc.response;
      
      if (std::is_base_of<Deferred, RPC>)
        server.m_Router.loop()->call(msg.reply(_response));
      else
        msg.send_reply(_response);
      
    }
  }

} // namespace llarp::rpc
```


-------------------------------

Updated on 2023-01-17 at 22:01:25 +0000
